"""Standalone script to produce Gantt Charts from AutoSubmit logs.

Example usage:

    $ python create_ganttchart.py o005 --chunks 38 52 --ignore AQUA

This produces the figure posted in:

    https://earth.bsc.es/gitlab/digital-twins/de_340-2/simulation-runs/-/issues/82#note_368241

Full options for plot creation:

    $ python create_ganttchart.py -h
"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter
from datetime import datetime
from pathlib import Path
import warnings

import matplotlib.pyplot as plt

from destine_analytics.logreport import get_aslogs_df, jobs_df_from_aslogs_df
from destine_analytics.plots import plot_job_gantt_chart as inner_plot_job_gantt_chart
from destine_analytics.decorators.asrun import asrun_with_jobs_available, asrun_with_stats
from destine_analytics.utils import get_available_expids
from destine_analytics.paths import LOG_DIR

# Prefixes of files to ignore in Gantt Chart by default
IGNORE_PREFIX = (".", "ls", "config", "githook", "mother_request", "only_lra")


def _get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="Gantt Chart plotter for Autosubmit logs.",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "expid",
        type=str,
        choices=get_available_expids(),
        help="Experiment id",
    )
    parser.add_argument(
        "--chunks",
        type=int,
        nargs=2,
        metavar=("START", "END"),
        default=(None, None),
        help="First and last chunks to plot.",
    )
    parser.add_argument(
        "--ignore",
        type=str,
        nargs="*",
        default=[],
        help="Ignore jobs starting with these arguments.",
    )
    parser.add_argument(
        "--color",
        choices=("job", "chunk", "status"),
        default="chunk",
        help="Whether to color jobs by job, by chunk or by status.",
    )
    parser.add_argument(
        "--starttype",
        choices=("started", "submitted"),
        default="started",
        help="Whether the start of the bars is at AS 'started' or 'submitted' status.",
    )
    parser.add_argument(
        "--cmap", type=str, default="tab20", help="Matplotlib colormap to use."
    )
    parser.add_argument(
        "--interval",
        type=int,
        default=-1,
        help="Interval for time ticks in x axis. -1 sets it to 'automatic'. Default: -1",
    )
    parser.add_argument(
        "--pivot",
        action="store_true",
        help="All bars are classified by chunk in the y axis.",
    )
    parser.add_argument(
        "--no-pivot",
        action="store_false",
        dest="autopivot",
        help="Prevent bars from being classified by chunk in the y axis, even if splits overlap.",
    )
    parser.add_argument(
        "--by-split",
        action="store_true",
        help="Gantt Chart entries are separated by split.",
    )
    parser.add_argument(
        "--by-member",
        action="store_true",
        help="Gantt Chart entries are separated by split.",
    )
    parser.add_argument(
        "--durations",
        action="store_true",
        help="Add the summed durations of all the jobs of each entry in the Gantt chart.",
    )
    parser.add_argument(
        "--no-edge",
        action="store_false",
        dest="edge",
        help="Add the summed durations of all the jobs of each entry in the Gantt chart.",
    )
    parser.add_argument(
        "--dpi",
        type=int,
        help="Add the summed durations of all the jobs of each entry in the Gantt chart.",
    )
    parser.add_argument(
        "--after",
        type=str,
        help="Date after which jobs are read. Format: YYYYMMDD or YYYYMMDDhhmmss",
    )
    parser.add_argument(
        "--y-pos",
        type=float,
        nargs="+",
        dest="y_positions",
        help="Relative positions of line info",
    )
    parser.add_argument(
        "--y-bar",
        type=float,
        default=0.4,
        dest="bar_height",
        help="Height of the bars. Default: 0.4",
    )
    parser.add_argument(
        "--y-factor",
        type=float,
        default=0.4,
        dest="height_factor",
        help="Factor to create height of the plot. Default: 0.4",
    )
    parser.set_defaults(autopivot=True)
    return parser


def _filter_by_chunk_interval(df, chunks):
    return df[(df["chunk"] >= chunks[0]) & (df["chunk"] <= chunks[1])]


def _parse_date_or_none(after: str | None):
    if after is None:
        return None
    if len(after) == 8:
        return datetime.strptime(after, r"%Y%m%d")
    if len(after) == 14:
        return datetime.strptime(after, r"%Y%m%d%H%M%S")
    raise ValueError("Date must be in format YYYYMMDD or YYYYYMMDDhhmmss.")


def _filter_by_date_min(df, date, starttype):
    if len(date) == 8:
        fmt = r"%Y%m%d"
    elif len(date) == 14:
        fmt = r"%Y%m%d%H%M%S"
    else:
        raise ValueError(
            "Date must be in format YYYYMMDD or YYYYYMMDDhhmmss."
        )
    return df[df[starttype] >= datetime.strptime(date, fmt)]
    
    
def main():
    args = _get_parser().parse_args()

    expid = args.expid
    chunks0, chunks1 = args.chunks
    ignore = args.ignore
    color = args.color
    starttype = args.starttype
    cmap = plt.get_cmap(args.cmap)
    interval = args.interval
    pivot = args.pivot
    autopivot = args.autopivot
    by_split = args.by_split
    by_member = args.by_member
    durations = args.durations
    edge = args.edge
    dpi = args.dpi
    after = _parse_date_or_none(args.after)
    y_positions = args.y_positions
    height_factor = args.height_factor
    bar_height = args.bar_height

    if by_member and not by_split:
        warnings.warn(
            "Independent pivots are not yet implemented. "
            "Therefore, --by-member has the same effect as --by-member --by-split."
        )

    print("Reading logs...")
    experiment_path = LOG_DIR / expid / "ASFILES"
    aslogs = get_aslogs_df(experiment_path, ignore_prefix=IGNORE_PREFIX)

    print("Extracting jobs...")
    jobs = jobs_df_from_aslogs_df(aslogs, experiment_path)

    print("Generating Gantt Chart plot...")
    plot_job_gantt_chart = asrun_with_stats(
        inner_plot_job_gantt_chart,
        log_path=(LOG_DIR / expid / "ASLOGS"),
        after=after,
        rel_y_position=y_positions,
        color_text="#006600",
    )

    if chunks0 is not None and chunks1 is not None:
        jobs = _filter_by_chunk_interval(jobs, (chunks0, chunks1))
        title = f"{expid} Simulation Timeline - Chunks [{chunks0}, {chunks1}]"
        filename = f"{expid}_simulation_timeline_{chunks0}_{chunks1}.png"
    else:
        title = f"{expid} Simulation Timeline"
        filename = f"{expid}_simulation_timeline.png"
    
    if after is not None:
        jobs = _filter_by_date_min(jobs, after, starttype)

    jobs = jobs[
        jobs["job"].apply(
            lambda x: not any(x.startswith(element) for element in ignore)
        )
    ]

    if after is not None:
        jobs = jobs[jobs[starttype] >= after]

    jobs = jobs[
        jobs["job"].apply(
            lambda x: not any(x.startswith(element) for element in ignore)
        )
    ]

    fig = plot_job_gantt_chart(
        jobs,
        title=title,
        color_key=color,
        starttype=starttype,
        cmap=cmap,
        interval=interval,
        pivot=(
            "member"
            if by_member
            else (
                "split"
                if by_split
                else "chunk" if pivot else "on_overlap" if autopivot else False
            )
        ),
        add_durations=durations,
        edge=edge,
        bar_height=bar_height,
        height_factor=height_factor,
    )
    Path("figures").mkdir(exist_ok=True)
    fig.savefig(f"figures/{filename}", dpi=dpi)
    print(f"Saved figures/{filename}")


if __name__ == "__main__":
    main()
